////////////////////////////////////////////////////////////////////////////////
//
// camera.cpp
// Methodes de la classe Camera
//
// Editeur de roller coaster
// Projet de realite virtuelle (INF6802)
// Herve Lombaert et Yves Chouinard
// Automne 2005
//
////////////////////////////////////////////////////////////////////////////////


#include "camera.h"


////////////////////////////////////////////////////////////////////////////////
//
// API de la classe Camera
//
// Camera::init()
//
// Cree et retourne un noeud de transformation OpenSG. Ce noeud est destine a
// etre ajoute au sommet du graphe de scene. La navigation est faite par la
// suite a l'aide des methodes setFirstPersonPosition() (mode premiere
// personne), translate() (mode en retrait) et toggleMode() qui change d'un
// mode a l'autre.
//
// Camera::setFirstPersonPosition()
//
// Change la transformation de la camera en mode premiere personne. On lui
// donne le parcours (path) et la position courante sur le parcours
// (walkingDistance).
//
// Camera::translate()
//
// Change la transformation de la camera en mode en retrait. On lui donne le
// deplacement en x et en y. L'axe des z represente la verticale.
//
////////////////////////////////////////////////////////////////////////////////


osg::NodePtr Camera::init()
{
	// Initial off camera transform
	offMatrix.setTranslate(osg::Vec3f(0, 0, -30));
	offMatrix.setRotate(osg::Quaternion(osg::Vec3f(1, 0, 0), -3.14159*75/180));

	// Camera transform - matrix set by calcTransform()
	transform = osg::Transform::create();
	osg::addRefCP(transform);

	// Camera node
	osg::NodePtr node = osg::Node::create();
	osg::beginEditCP(node);
	node->setCore(transform);
	osg::addRefCP(node);
	osg::endEditCP(node);

	// Off camera mode by default
	mode = OFF_CAMERA_MODE;
	calcTransform();

	return node;
}


void Camera::setFirstPersonPosition(Path &path, float walkingDistance)
{
	this->path = path;
	this->walkingDistance = walkingDistance;
	calcTransform();
}


void Camera::translate(float dx, float dy)
{
	offMatrix[3][0] += dx;
	offMatrix[3][2] += dy;
	calcTransform();
}


////////////////////////////////////////////////////////////////////////////////
//
// Camera::calcTransform()
//
// Methode privee qui met a jour la transformation OpenSG a partir de la
// position sur le parcours (mode premiere personne) ou de la matrice
// actuelle du mode en retrait.
//
// Utilisee par init(), setFirstPersonPosition(), translate() et
// toggleMode().
//
////////////////////////////////////////////////////////////////////////////////


void Camera::calcTransform()
{
	switch (mode)
	{
		case OFF_CAMERA_MODE:
		{
			transform->setMatrix(offMatrix);
		}
		break;

		case FIRST_PERSON_CAMERA_MODE:
		{
			osg::Matrix m1, m2;

			// For -20<=x<=20 and -20<=y<=20, the structure looks a bit small
			m1.setTranslate(osg::Vec3f(0, -2, 0));
			m1.setScale(4);

			// Put z-axis up instead of pointing in the back
			m2.setRotate(osg::Quaternion(osg::Vec3f(1, 0, 0), -3.14159/2));
			m1.mult(m2);
			//m1.setIdentity();

			// World origin (CAVE center floor) in first wagon
			float m[16];
			path.interpolate(walkingDistance, m);
			m2.setValue(m);
			m2.invert();
			m1.mult(m2);

			transform->setMatrix(m1);
		}
		break;
	}
}
