////////////////////////////////////////////////////////////////////////////////
//
// cursor.cpp
// Methodes de la classe Cursor
//
// Editeur de roller coaster
// Projet de realite virtuelle (INF6802)
// Herve Lombaert et Yves Chouinard
// Automne 2005
//
////////////////////////////////////////////////////////////////////////////////


#include "cursor.h"
#include <OpenSG/OSGTransform.h>


static float cursorscale = 9.0f;
static float cursorx = 0;
static float cursory = 0;
static float cursorz = -3.0;


////////////////////////////////////////////////////////////////////////////////
//
// Cursor::init()
//
// Cree le curseur et retourne son noeud OpenSG, qui est destine a etre
// ajoute au graphe de scene. Par defaut, le curseur est un gros point rouge.
//
////////////////////////////////////////////////////////////////////////////////


osg::NodePtr Cursor::init()
{
	osg::SimpleMaterialPtr mat = osg::SimpleMaterial::create();
	beginEditCP(mat);
	  mat->setDiffuse(osg::Color3f(1,0,0));
	  mat->setAmbient(osg::Color3f(0.4,0.4,0.4));
	  mat->setTransparency(0.3);
	endEditCP(mat);

	geometry = osg::makeSphere(2, 1.5);
	beginEditCP(geometry);
	osg::GeometryPtr::dcast(geometry->getCore())->setMaterial(mat);
	beginEditCP(geometry);

	osg::TransformPtr t = osg::Transform::create();
	transf = osg::Node::create();
	beginEditCP(transf);
	  transf->setCore(t);
	  transf->addChild(geometry);
	endEditCP(transf);

	return transf;
}


////////////////////////////////////////////////////////////////////////////////
//
// Cursor::setColor
// Pour le monde fancy qui n'aiment pas les curseurs rouges.
//
////////////////////////////////////////////////////////////////////////////////


void Cursor::setColor(float r, float g, float b, float a)
{
	osg::GeometryPtr geo = osg::GeometryPtr::dcast(geometry->getCore());
	osg::SimpleMaterialPtr mat = osg::SimpleMaterialPtr::dcast(geo->getMaterial());
	beginEditCP(mat);
	  mat->setDiffuse(osg::Color3f(r,g,b));
	  mat->setTransparency(a);
	endEditCP(mat);
}


////////////////////////////////////////////////////////////////////////////////
//
// Cursor::setMatrix()
//
// C'est avec cette methode que se fait le deplacement du curseur. On lui
// donne simplement la matrice de la wand. C'est la methode qui s'occupe de
// faire la transformation d'echelle pour que le curseur se deplace plus que
// la wand.
//
////////////////////////////////////////////////////////////////////////////////


void Cursor::setMatrix(osg::Matrix mat)
{
	osg::TransformPtr t = osg::TransformPtr::dcast(transf->getCore());
	osg::Matrix m=mat; m.setTranslate(cursorscale*(mat[3][0]+cursorx),
					  cursorscale*(-mat[3][2]+cursory),
					  cursorscale*(mat[3][1]+cursorz));
	beginEditCP(t);
	  t->setMatrix(m);
	endEditCP(t);
}


////////////////////////////////////////////////////////////////////////////////
//
// Cursor::pick()
//
// Pour selectionner un point de controle. On specifie le parcours avec le
// parametre path. Retourne le point de controle le plus pres de la position
// courante du curseur. L'une des deux versions donne aussi la distance entre
// les deux via le parametre distSquared. L'application peut decider de ne
// selectionner aucun point si la distance est trop grande.
//
////////////////////////////////////////////////////////////////////////////////


Path::PointID Cursor::pick(Path &path)
{
	osg::TransformPtr t = osg::TransformPtr::dcast(transf->getCore());
	osg::Matrix m = t->getMatrix();
	float x = m[3][0];
	float y = m[3][1];
	float z = m[3][2];

	Path::PointID id = path.pickControlPoint(x,y,z);

	return id;
}


Path::PointID Cursor::pick(Path &path, float &distSquared)
{
	// cursor position
	osg::TransformPtr t = osg::TransformPtr::dcast(transf->getCore());
	osg::Matrix m = t->getMatrix();
	float x = m[3][0];
	float y = m[3][1];
	float z = m[3][2];

	// closest control point
	Path::PointID id = path.pickControlPoint(x,y,z);

	// distance to it
	float matrix[16];
	path.getControlPoint(id, matrix);
	distSquared =   (matrix[12]-x)*(matrix[12]-x) +
			(matrix[13]-y)*(matrix[13]-y) +
			(matrix[14]-z)*(matrix[14]-z);

	return id;
}


////////////////////////////////////////////////////////////////////////////////
//
// Cursor::getTranslate()
//
// La position courante du curseur est utilisee pour modifier de facon
// correspondante celle du point de controle selectionne en mode drag and
// drop.
//
////////////////////////////////////////////////////////////////////////////////


void Cursor::getTranslate(float &x, float &y, float &z)
{
	osg::TransformPtr t = osg::TransformPtr::dcast(transf->getCore());
	osg::Matrix m = t->getMatrix();
	x = m[3][0];
	y = m[3][1];
	z = m[3][2];
}


////////////////////////////////////////////////////////////////////////////////
//
// Cursor::show()
// Cursor::hide()
// Cursor::toggleVisibility()
//
// Pour que le curseur soit visible en mode d'edition, mais cache en mode de
// simulation.
//
////////////////////////////////////////////////////////////////////////////////


void Cursor::show()
{
	osg::GeometryPtr geo = osg::GeometryPtr::dcast(geometry->getCore());
	osg::SimpleMaterialPtr mat = osg::SimpleMaterialPtr::dcast(geo->getMaterial());
	beginEditCP(mat);
	  mat->setTransparency(0.3);
	endEditCP(mat);
}


void Cursor::hide()
{
	osg::GeometryPtr geo = osg::GeometryPtr::dcast(geometry->getCore());
	osg::SimpleMaterialPtr mat = osg::SimpleMaterialPtr::dcast(geo->getMaterial());
	beginEditCP(mat);
	  mat->setTransparency(1.0);
	endEditCP(mat);
}


void Cursor::toggleVisibility()
{
	osg::GeometryPtr geo = osg::GeometryPtr::dcast(geometry->getCore());
	osg::SimpleMaterialPtr mat = osg::SimpleMaterialPtr::dcast(geo->getMaterial());

	// toggle transparency
	float transparency = mat->getTransparency();
	if (transparency) transparency = 0.3; else transparency = 1.0;
	printf("visibility: %f\n", transparency);

	beginEditCP(mat);
	  mat->setTransparency(transparency);
	endEditCP(mat);

	beginEditCP(geo);
	  geo->setMaterial(mat);
	endEditCP(geo);
}
