/**
 * @file path.cpp
 * @author Herve Lombaert
 * @date October 16th, 2005
 *
 * Create path from control points and interpolate a position and orientation
 * on path
 *
 * Positions are interpolated using the Catmull-Rom equation between
 * 4 control points
 * The right vector is the cross product of the last and the current
 * direction vectors
 * The up vector is the cross product of the right and the direction vectors
 */

#include <cstdio>
#include <cmath>
#include "path.h"

/** @brief	give a new control point position **/
Path::PointID	Path::addControlPoint(float x, float y, float z)
{
	try
	{
		controlPoints.push_back(ControlPoint(x,y,z));
		Path::PointID id = --(controlPoints.end());
		updateRotationFrom(id);
		return id;
	}
	catch(...) { throw Exception(__FILE__, __LINE__, "addControlPoint(), unknown error"); }
}

/** @brief	insert a control point position
    @note	find the closest control point and insert new one after it **/
Path::PointID	Path::insertControlPoint(float x, float y, float z)
{
	try
	{
		PointID id = pickControlPoint(x,y,z);
		if(id != controlPoints.begin())
		{
			PointID p0Id, p1Id, p2Id, p3Id;

			// check if we should insert the new point after the closest point
			PointID nextid = ++(PointID(id));
			getNeighbors(nextid, p0Id, p1Id, p2Id, p3Id);
			float nextMinDist = 1e+4;
			for(float t=0; t<1; t+=0.1)
			{
				vec4 p = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, t);
				float dist = (p[0]-x)*(p[0]-x) +
					     (p[1]-y)*(p[1]-y) +
					     (p[2]-z)*(p[2]-z);
				if(dist<nextMinDist)
					nextMinDist = dist;
			}

			// check if we should insert the new point after the closest point
			getNeighbors(id, p0Id, p1Id, p2Id, p3Id);
			float prevMinDist = 1e+4;
			for(float t=0; t<1; t+=0.1)
			{
				vec4 p = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, t);
				float dist = (p[0]-x)*(p[0]-x) +
					     (p[1]-y)*(p[1]-y) +
					     (p[2]-z)*(p[2]-z);
				if(dist<prevMinDist)
					prevMinDist = dist;
			}

			// if closest point on path is after selected point, insert after it
			if(nextMinDist < prevMinDist)
				id = nextid; // new point is after picked control point
		}
		else
		{
			++id; // place new control point after beginning
		}
		id = controlPoints.insert(id, ControlPoint(x,y,z));
		updateRotationFrom(id);
		return id;
	}
	catch(...) { throw Exception(__FILE__, __LINE__, "insertControlPoint(), unknown error"); }
}

/** @brief	remove a control point **/
void	Path::removeControlPoint(Path::PointID pointId)
{
	try
	{
		if(pointId==controlPoints.end())
			throw Exception(__FILE__, __LINE__, "removeControlPoints(), out of bound");

		pointId = controlPoints.erase(pointId);
		updateRotationFrom(pointId);
	}
	catch(Exception &e) { printf("from removeControlPoints()\n"); throw e; }
	catch(...) { throw Exception(__FILE__, __LINE__, "removeControlPoints(), unknown error"); }
}

/** @brief	change a control point position **/
void	Path::changeControlPoint(Path::PointID pointId, float x, float y, float z)
{
	try
	{
		if(pointId==controlPoints.end())
			throw Exception(__FILE__, __LINE__, "changeControlPoint(), out of bound");

		pointId->pos[0] = x;
		pointId->pos[1] = y;
		pointId->pos[2] = z;

		updateRotationFrom(pointId);
	}
	catch(Exception &e) { printf("from changeControlPoint()\n"); throw e; }
	catch(...) { throw Exception(__FILE__, __LINE__, "changeControlPoint(), unknown error"); }
}

/** @brief	interpolate a transformation matrix on path
    @note	4x4 column wise matrix **/
void	Path::interpolate(float length, float* matrix) const
{
	try
	{
		/** variable:	a certain number of control points at the end
						can have the up vector pointing up,
						this is a useful if the track has to end with
						the same orientation as the beginning **/
		const int maxDistToEnd = 4;
		
		if(controlPoints.empty())
			throw Exception(__FILE__, __LINE__, "interpolate(), no control points");
		
		/** get previous and current control point rotation **/
		std::vector<ControlPoint>::const_iterator p0Id, p1Id, p2Id, p3Id; float time;
		float l = 0;

		l = (length-0.01<0) ? 0 : length-0.01;
		getNeighbors(l, p0Id, p1Id, p2Id, p3Id, time); // before
		vec4 p0 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, time);
		
		l = (length+0.01>totalLength) ? totalLength : length+0.01;
		getNeighbors(l, p0Id, p1Id, p2Id, p3Id, time); // after
		vec4 p2 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, time);

		/** interpolate position **/
		getNeighbors(length, p0Id, p1Id, p2Id, p3Id, time); // now
		vec4 p1 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, time);

		/** interpolate up vector **/
		vec4 up[4] = {	quatToRotationMatrix(p0Id->rot)[2],
						quatToRotationMatrix(p1Id->rot)[2],
						quatToRotationMatrix(p2Id->rot)[2],
						quatToRotationMatrix(p3Id->rot)[2]	};

		/** make last vector rotation similar to default rotation **/
		int distToEnd = 0;
		std::vector<ControlPoint>::const_iterator pId = p3Id;
		while(pId != controlPoints.end() && distToEnd<maxDistToEnd)
			++distToEnd, ++pId;
		for(int i=0; i<maxDistToEnd-distToEnd && i<4; ++i)
			up[3-i] = vec4(0,0,1,1);
				
		/** construct rotation matrix, keep dir, interpole up **/
		mtx44 rotMat;
		rotMat[1] = p2-p0;						/* dir */		rotMat[1] /= norm(rotMat[1]);
		rotMat[2] = sampleCatmullRom(up[0],up[1],up[2],up[3],time);	rotMat[2] /= norm(rotMat[2]);
		rotMat[0] = cross(rotMat[1],rotMat[2]);	/* right */		rotMat[0] /= norm(rotMat[0]);
				
		/** fill in matrix **/
		matrix[0] = rotMat[0][0];
		matrix[1] = rotMat[0][1];
		matrix[2] = rotMat[0][2];
		matrix[3] = 0;

		matrix[4] = rotMat[1][0];
		matrix[5] = rotMat[1][1];
		matrix[6] = rotMat[1][2];
		matrix[7] = 0;
		
		matrix[8] = rotMat[2][0];
		matrix[9] = rotMat[2][1];
		matrix[10] = rotMat[2][2];
		matrix[11] = 0;
		
		matrix[12] = p1[0];
		matrix[13] = p1[1];
		matrix[14] = p1[2];
		matrix[15] = 1;
	}
	catch(Exception &e) { printf("from interpolate()\n"); throw e; }
	catch(...) { throw Exception(__FILE__, __LINE__, "interpolate(), unknown error"); }
}

/** @brief	get total length of path **/
float	Path::getTotalLength() const
{
	try
	{
		return totalLength;
	}
	catch(...) { throw Exception(__FILE__, __LINE__, "getTotalLength(), unknown error"); }
}

/** @brief	pick the closest control point to x,y,z **/
Path::PointID	Path::pickControlPoint(float x, float y, float z)
{
	try
	{
		/** test first point **/
		PointID pointId = controlPoints.begin();
		float distSquared = (pointId->pos[0]-x)*(pointId->pos[0]-x) + (pointId->pos[1]-y)*(pointId->pos[1]-y) + (pointId->pos[2]-z)*(pointId->pos[2]-z);

		float minDistSquared = distSquared;
		Path::PointID closestPointId = pointId;

		/** test all other points **/
		++pointId; // first point is already tested
		while(pointId != controlPoints.end())
		{
			distSquared = (pointId->pos[0]-x)*(pointId->pos[0]-x) + (pointId->pos[1]-y)*(pointId->pos[1]-y) + (pointId->pos[2]-z)*(pointId->pos[2]-z);
			if(distSquared < minDistSquared)
			{
				minDistSquared = distSquared;
				closestPointId = pointId;
			}
	
			++pointId;
		}

		return closestPointId;
	}
	catch(...) { throw Exception(__FILE__, __LINE__, "pickControlPoint(), unknown error"); }
}

/** @brief	return the id of the first control point **/
Path::PointID	Path::begin()
{
	try
	{
		return controlPoints.begin();
	}
	catch(...) { throw Exception(__FILE__, __LINE__, "begin(), unknown error"); }
}

/** @brief	return the first invalid id after the last control point **/
Path::PointID	Path::end()
{
	try
	{
		return controlPoints.end();
	}
	catch(...) { throw Exception(__FILE__, __LINE__, "end(), unknown error"); }
}

/** @brief	get the transformation matrix of the control point
    @note	4x4 column wise matrix **/
void	Path::getControlPoint(Path::PointID pointId, float* matrix) const
{
	try
	{
		if(controlPoints.empty())
			throw Exception(__FILE__, __LINE__, "getControlPointTransformation(), no control points");

		/** interpolate position **/
		vec4 pos = pointId->pos;
		
		/** interpolate rotation **/
		mtx44 rotMat = quatToRotationMatrix(pointId->rot);
		
		/** fill in matrix **/
		matrix[0] = rotMat[0][0];
		matrix[1] = rotMat[0][1];
		matrix[2] = rotMat[0][2];
		matrix[3] = 0;
		
		matrix[4] = rotMat[1][0];
		matrix[5] = rotMat[1][1];
		matrix[6] = rotMat[1][2];
		matrix[7] = 0;
		
		matrix[8] = rotMat[2][0];
		matrix[9] = rotMat[2][1];
		matrix[10] = rotMat[2][2];
		matrix[11] = 0;
		
		matrix[12] = pos[0];
		matrix[13] = pos[1];
		matrix[14] = pos[2];
		matrix[15] = 1;		
	}
	catch(Exception &e) { printf("from getControlPointTransformation()\n"); throw e; }
	catch(...) { throw Exception(__FILE__, __LINE__, "getControlPointTransformation(), unknown error"); }
}

/** @brief	constructor **/
Path::Path()
:	totalLength(0)
{
}

/** @brief	update the length and orientation of segments affected by a change of control point
    @note	segments affected by a control points are the 2 segments
		behind control points and 2 segment after control point **/
void	Path::updateRotationFrom(Path::PointID pointId)
{
	try
	{
		if(controlPoints.empty()) throw Exception(__FILE__, __LINE__, "updateSegmentAtPoint(), updating empty list");

		/** update neighbors **/
		Path::PointID p0Id, p1Id, p2Id, p3Id;
		getNeighbors(pointId, p0Id, p1Id, p2Id, p3Id);

		updateRotationAt(p0Id);
		updateRotationAt(p1Id);
		updateRotationAt(p2Id);
		updateRotationAt(p3Id);

		/** get total length **/
		std::vector<ControlPoint>::const_iterator pointId = controlPoints.begin();
		totalLength = 0;
		while(pointId != controlPoints.end())
		{
			totalLength += pointId->dist;
			++pointId;
		}
	}
	catch(Exception &e) { printf("from updateRotationAt()\n"); throw e; }
	catch(...) { throw Exception(__FILE__, __LINE__, "updateSegmentAtPoint(), unknown error"); }
}
	
/** @brief	update the segment length and orientation
    @param	pointId		id of the ending point of the segment
    @note	a segment is between two control points
		to know the new orientation, we use the previous right vector
		and compute the new up vector, and we use the previous up vector
		and compute the new right vector. We keep the best value (either
		the new up or the new right vector is larger) **/
void	Path::updateRotationAt(PointID pointId)
{
	try
	{
//////////////////////////////////////////////
		/** variable:	number of subsegment in a segment
		the segment length is the sum of all subsegment length **/
		const float timeStep = 0.1;

		if(pointId==controlPoints.end())
			throw Exception(__FILE__, __LINE__, "updateSegment(), out of bound");

		/** if it's the only point in list **/
		if(pointId == controlPoints.begin() && controlPoints.begin() == --(controlPoints.end()))
		{
			mtx44 rotMat(vec4(0,-1,0,0),vec4(1,0,0,0),vec4(0,0,1,0),vec4(0,0,0,1));
			pointId->rot = quatFromRotationMatrix(rotMat);
			return;
		}
		
		/** if it's the first control point set default values **/
		if(pointId == controlPoints.begin())
		{
			vec4 p1 = pointId->pos;
			
			PointID p0Id, p1Id, p2Id, p3Id;
			getNeighbors(++(PointID(pointId)), p0Id, p1Id, p2Id, p3Id);
			vec4 p2 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, timeStep);
			
			vec4 dir = p2-p1;				dir   /= norm(dir);
			vec4 up(0,0,1,1);
			vec4 right = cross(dir,up);		right /= norm(right);
			up = cross(right,dir);			up    /= norm(up);	// update right
			
			mtx44 rotMat(right,dir,up,vec4(0,0,0,1));
			pointId->rot = quatFromRotationMatrix(rotMat);
			return;
		}
		
		/** if it's not the first point, walk from last control point and update rotation **/
		PointID p0Id, p1Id, p2Id, p3Id;
		getNeighbors(pointId, p0Id, p1Id, p2Id, p3Id);
		
		mtx44 prevRotMat = quatToRotationMatrix(p1Id->rot);
		vec4 prevUp = prevRotMat[2];
		float walkingDistance = 0;
		
		/** walk on segment, update rotation **/
		for(float s=timeStep; s<1; s+=timeStep)
		{
			/** interpolate new position on segment **/
			vec4 p0 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, s-timeStep); // t=i-1
			vec4 p1 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, s);			// t=i
			vec4 p2 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, s+timeStep); // t=i+1
			
			/** update walking distance **/
			walkingDistance += norm(p1-p0); if(s+timeStep>=1) walkingDistance += norm(p2-p1);
			
			/** update rotation **/
			vec4 dir = p2-p0;				dir   /= norm(dir);
			vec4 up = prevUp;				up	  /= norm(up);
			// add roll to up vector here
			vec4 right = cross(dir,up);		right /= norm(right);
			up = cross(right,dir);			up    /= norm(up);
			
			/** go to the next point **/
			prevUp = up;
		}
		
		/** update distance to last control point **/
		pointId->dist = walkingDistance;
		
		/** get dir vector on control point **/
		vec4 p0,p2;
		
		p0 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, 1-0.01/p2Id->dist);
		
		if(p3Id == p2Id) // no control point after this one
			p2 = pointId->pos;
		else
		{
			getNeighbors(p3Id, p0Id, p1Id, p2Id, p3Id);
			p2 = sampleCatmullRom(p0Id->pos, p1Id->pos, p2Id->pos, p3Id->pos, (p2Id->dist)?0.01/p2Id->dist:0.01);
		}
		
		vec4 dir = p2-p0;				dir   /= norm(dir);
		vec4 up = prevUp;				up    /= norm(up);
		vec4 right = cross(dir,up);		right /= norm(right);
		up = cross(right,dir);			up    /= norm(up);
		
		/** set rotation of control point **/
		mtx44 rotMat(right,dir,up,vec4(0,0,0,1));
		pointId->rot = quatFromRotationMatrix(rotMat);
		
		/** always allow the shortest path during interpolation **/
		if(dot((--(PointID(pointId)))->rot,pointId->rot) < 0)
		   pointId->rot = -pointId->rot; // negation results in the same rotation
	}
	catch(Exception &e) { printf("from updateSegment()\n"); throw e; }
	catch(...) { throw Exception(__FILE__, __LINE__, "updateSegment(), unknown error"); }
}


/** @brief	get 4 neighboring points of a control point **/
void	Path::getNeighbors(std::vector<ControlPoint>::const_iterator pointId,
						   std::vector<ControlPoint>::const_iterator &p0Id,
						   std::vector<ControlPoint>::const_iterator &p1Id,
						   std::vector<ControlPoint>::const_iterator &p2Id,
						   std::vector<ControlPoint>::const_iterator &p3Id) const
{
	try
	{
		/** end of segment **/
		p2Id = pointId;
		
		/** point after segment **/
		p3Id = ++(std::vector<ControlPoint>::const_iterator (p2Id));
		if(p3Id==controlPoints.end()) p3Id = p2Id;
		
		/** start of segment **/
		if(pointId==controlPoints.begin())	p1Id = p2Id;
		else					p1Id = --(std::vector<ControlPoint>::const_iterator (p2Id));
		
		/** point before segment **/
		if(p1Id==controlPoints.begin())		p0Id = p1Id;
		else					p0Id = --(std::vector<ControlPoint>::const_iterator (p1Id));
	}
	catch(...) { throw Exception(__FILE__, __LINE__, "getNeighbors(), unknown error"); }
}


/** @brief	get 4 neighboring points at a given distance **/
void	Path::getNeighbors(	float length,
							std::vector<ControlPoint>::const_iterator &p0Id,
							std::vector<ControlPoint>::const_iterator &p1Id,
							std::vector<ControlPoint>::const_iterator &p2Id,
							std::vector<ControlPoint>::const_iterator &p3Id,
							float &time) const
{
	try
	{
		/** walkingDistance is the distance from start to the control point right before
		    our interpolated point, walkingPointId points to the controlPoint after it
		 **/
		if(controlPoints.size()<2)
			throw Exception(__FILE__, __LINE__, "getNeighbors(), need 2 points");
		
		/** get the segment in which length points at **/
		std::vector<ControlPoint>::const_iterator walkingPointId = ++(controlPoints.begin());
		float walkingDistance = 0;

		/** if length is slightly larger than total length (floating precision imperfection)
		    return last segment **/
		if(length>=totalLength && length-totalLength<0.01)
		{
			walkingPointId = --(controlPoints.end());
			getNeighbors(walkingPointId, p0Id, p1Id, p2Id, p3Id);
			time = 1.0f;
			return;
		}

		/** walk on path until fniding our segment **/
		while(walkingDistance+walkingPointId->dist < length)
		{
			if(walkingPointId == controlPoints.end())
				throw Exception(__FILE__, __LINE__, "getNeighbors(), out of bound");

			walkingDistance += walkingPointId->dist;
			++walkingPointId; // now points at the end of the next segment
		}
/** get neighbors **/
		getNeighbors(walkingPointId, p0Id, p1Id, p2Id, p3Id);
		
		/** compute time (0..1) **/
		time = (length-walkingDistance) / walkingPointId->dist;
	}
	catch(Exception &e) { printf("from getNeighbors()\n"); throw e; }
	catch(...) { throw Exception(__FILE__, __LINE__, "getNeighbors(), unknown error"); }
}

/** @brief	get 4 neighboring points of a control point **/
void	Path::getNeighbors(PointID pointId,
						   PointID &p0Id,
						   PointID &p1Id,
						   PointID &p2Id,
						   PointID &p3Id)
{
	try
	{
		/** end of segment **/
		p2Id = pointId;
		
		/** point after segment **/
		p3Id = ++(PointID(p2Id));
		if(p3Id==controlPoints.end()) p3Id = p2Id;
		
		/** start of segment **/
		if(pointId==controlPoints.begin())	p1Id = p2Id;
		else					p1Id = --(PointID(p2Id));
		
		/** point before segment **/
		if(p1Id==controlPoints.begin())		p0Id = p1Id;
		else					p0Id = --(PointID(p1Id));
	}
	catch(...) { throw Exception(__FILE__, __LINE__, "getNeighbors(), unknown error"); }
}

/** @brief	io **/
void	Path::save(const char* filename) const
{
	FILE *fd = fopen(filename, "w");
	if(fd)
	{
		std::vector<Path::ControlPoint>::const_iterator pb = controlPoints.begin(), pe = controlPoints.end();
		while(pb != pe) {
			fprintf(fd, "%f %f %f\n", pb->pos[0], pb->pos[1], pb->pos[2]);
			++pb;
		}
		fclose(fd);
	} else { printf("problem when opening: %s\n", filename); }
}

/** @brief	io **/
bool	Path::load(const char* filename)
{
	FILE *fd = fopen(filename, "r");
	if(fd)
	{
		float x,y,z;
		int n = fscanf(fd, "%f %f %f", &x, &y, &z);
		while(n>0)
		{
			addControlPoint(x,y,z);
			n = fscanf(fd, "%f %f %f\n", &x, &y, &z);
		}
		fclose(fd);
		return true;
	} else { printf("problem when opening: %s\n", filename); return false; }
}

Path::Exception::Exception(const char* file, int line, const char* message)
:	line(line)
{
	strncpy(this->file, file, 512);
	strncpy(this->message, message, 512);
}

void Path::Exception::what() const
{
	printf("Path Exception in file %s at line %d: %s\n", file, line, message);
}
