/**
 * @file path.h
 * @author Herve Lombaert
 * @dataOctober 16th, 2005
 *
 * Create path from control points and interpolate a position and orientation
 * on path
 *
 * Positions are interpolated using the Catmull-Rom equation between
 * 4 control points
 * The right vector is the cross product of the last and the current
 * direction vectors
 * The up vector is the cross product of the right and the direction vectors
 **/

#ifndef PATH_H
#define PATH_H

#include <vector>
#include "vquats.h"

/** @class	Interpolate a position on a path **/
class Path
{
private:
	class ControlPoint {
	public:		ControlPoint(float x, float y, float z) : pos(x,y,z,1), rot(0,0,0,1), dist(0) {}
				vec4 pos;	/** position of control point (can be modified) **/
				quat rot;	/** rotation of control point (updated automatically) **/
				float dist; /** distance to last control point **/
	};

public:
	typedef std::vector<ControlPoint>::iterator PointID; /** Control point ID **/

	/** @class	Exception **/
	class Exception {
	public:		Exception(const char* file, int line, const char* message);
				void	what() const;
	private:	char file[512]; int line; char message[512];
	};

public:
	/** @brief	give a new control point position **/
	PointID	addControlPoint(float x, float y, float z);

	/** @brief	insert a control point position **/
	PointID	insertControlPoint(float x, float y, float z);

	/** @brief	remove a control point **/
	void	removeControlPoint(PointID pointId);

	/** @brief	change a control point position **/
	void	changeControlPoint(PointID pointId, float x, float y, float z);

	/** @brief	interpolate a transformation matrix on path
	    @note	4x4 column wise matrix **/
	void	interpolate(float length, float matrix[16]) const;

	/** @brief	get total length of path **/
	float	getTotalLength() const;

	/** @brief	pick the closest control point to x,y,z **/
	PointID	pickControlPoint(float x, float y, float z);

	/** @brief	return the id of the first control point **/
	PointID	begin();

	/** @brief	return the first invalid id after the last control point **/
	PointID	end();

	/** @brief	get the transformation matrix of the control point
	    @note	4x4 column wise matrix **/
	void	getControlPoint(PointID pointId, float* matrix) const;

	/** @brief	io **/
	void	save(const char* filename) const;
	bool	load(const char* filename);

	/** @brief	constructor **/
	Path();

private:
	std::vector<ControlPoint>	controlPoints;	/** vector of control points **/
	float	totalLength;	/** total length of path **/

	/** @brief	get 4 neighboring points of a control point **/
	void	getNeighbors(	std::vector<ControlPoint>::const_iterator  pointId,
							std::vector<ControlPoint>::const_iterator  &p0Id,
							std::vector<ControlPoint>::const_iterator  &p1Id,
							std::vector<ControlPoint>::const_iterator  &p2Id,
							std::vector<ControlPoint>::const_iterator  &p3Id) const;
	void	getNeighbors(	PointID pointId,
							PointID &p0Id,
							PointID &p1Id,
							PointID &p2Id,
							PointID &p3Id);
	
	/** @brief	get 4 neighboring points at a given distance **/
	void	getNeighbors(	float length,
							std::vector<ControlPoint>::const_iterator &p0Id,
							std::vector<ControlPoint>::const_iterator &p1Id,
							std::vector<ControlPoint>::const_iterator &p2Id,
							std::vector<ControlPoint>::const_iterator &p3Id,
							float &time) const;
	
	/** @brief	update the length and rotation of segments affected by a control point
	    @note	segments affected by a control points are the 2 segments
			behind control points and 2 segment after control point **/
	void	updateRotationFrom(PointID pointId);

	/** @brief	update the length and rotation of segment
	    @param	pointId		id of the ending point of the segment
	    @note	a segment is between two control points **/
	void	updateRotationAt(PointID pointId);
};

#endif
