////////////////////////////////////////////////////////////////////////////////
//
// rc.cpp
// Methodes de la classe RC
//
// Editeur de roller coaster
// Projet de realite virtuelle (INF6802)
// Herve Lombaert et Yves Chouinard
// Automne 2005
//
////////////////////////////////////////////////////////////////////////////////


#include <cstdlib>
#include <ctime>

#include <GL/gl.h>
#include <gmtl/Matrix.h>
#include <gmtl/External/OpenSGConvert.h>

#if SOUND
#include <snx/sonix.h>
#include <snx/SoundHandle.h>
#endif

#include <OpenSG/OSGDirectionalLight.h>
#include <OpenSG/OSGTransform.h>
#if TEXTURE
#include <OpenSG/OSGImage.h>
#include <OpenSG/OSGSimpleTexturedMaterial.h>
#endif

#include "rc.h"


////////////////////////////////////////////////////////////////////////////////
//
// RC::initScene()
//
// Methode appelee par VR Juggler au debut de l'execution. Elle initialise
// les objets de VR Juggler, initialise ses propres variables et construit le
// graphe de scene.
//
////////////////////////////////////////////////////////////////////////////////


void RC::initScene()
{
	mode = NONE;

	// VR Juggler
	mHead.init("VJHead");
	mWand.init("VJWand");
	mButton0.init("VJButton0");
	mButton1.init("VJButton1");
	mButton2.init("VJButton2");
	mLastTiming.secd(0);

	// Path
	if (!path.load("current.path"))
	{
		path.addControlPoint(-14,0,1);
		path.addControlPoint(-13,0,1);
		path.addControlPoint(-10,0,1);
		path.addControlPoint(0,0,3);
		path.addControlPoint(5,0,5);
		path.addControlPoint(8,2,2);
		path.addControlPoint(12,7,6);
		path.addControlPoint(10,5,9);
		path.addControlPoint(9,1,5);
		path.addControlPoint(12,-5,4);
		path.addControlPoint(7,-8,10);
		path.addControlPoint(3,-12,5);
		path.addControlPoint(-5,-10,1);
		path.addControlPoint(-18,-2,3);
		path.addControlPoint(-15,0,1);
		path.addControlPoint(-14,0,1);
	}

	// Speed
	speed = new Speed(&path);

	// Structure
	osg::NodePtr structureNode = structure.init();
	structure.setPath(path);

	// Train
	osg::NodePtr trainNode = train.init(4);
	train.setPosition(path, 0);

	// Ground
	osg::NodePtr groundNode = ground.init();

	// Light
	osg::NodePtr light = initLight();
	osg::beginEditCP(scene);
	  light->addChild(groundNode);
	  light->addChild(initSky());
	  light->addChild(initHouse());
	  light->addChild(structureNode);
	  light->addChild(trainNode);
	osg::endEditCP(scene);

	// Camera (scene transform)
	osg::NodePtr cameraNode = camera.init();
	osg::beginEditCP(cameraNode);
	  cameraNode->addChild(light);
	  cameraNode->addChild(cursor.init());
	osg::endEditCP(cameraNode);

	// Top node
	osg::TransformPtr t = osg::Transform::create();
	scene = osg::Node::create();
	osg::beginEditCP(scene);
	  scene->setCore(t);
	  scene->addChild(cameraNode);
	osg::endEditCP(scene);

#if 0
	camera.toggleMode();
#endif

#if SOUND
      // Play a sound
      {
        snx::sonix::instance()->changeAPI("OpenAL");

        snx::SoundInfo si;
        //si.filename = "sounds/ocean_edge.wav";
	//si.filename = "sounds/seagull.wav";
	si.filename = "sounds/train.wav";
        si.datasource = snx::SoundInfo::FILESYSTEM;
	si.volume = 1;

	snx::SoundHandle background("background");
	background.configure(si);
	background.trigger(-1); // loop forever

	// Sound to be triggered when the train gains speed
	si.filename = "sounds/roller_coaster.wav";
	si.datasource = snx::SoundInfo::FILESYSTEM;
	si.volume = 0.1;

	soundSpeed.init("speed");
	soundSpeed.configure(si);
      }
#endif

	// init random seed
	srand((int)time(NULL));
}


////////////////////////////////////////////////////////////////////////////////
//
// RC::preFrame()
//
// Methode appelee par VR Juggler entre les affichages de frames. On y
// calcule la position courante du train, qui varie en fonction du temps, la
// position de la camera, qui suit le train en mode de simulation et on bouge
// le curseur en fonction de la position de la wand.
//
// On y traite aussi les boutons presses. Le bouton 1 sert a creer ou
// modifier les points de controles. Le bouton 2 sert a effacer les points de
// controle et a naviguer (selon l'orientation de la wand) en mode d'edition,
// selon que le curseur est suffisamment pres ou non d'un point de controle.
// Le bouton de droite sert a passer du mode d'edition au mode de simulation
// et vice-versa.
//
////////////////////////////////////////////////////////////////////////////////


void RC::preFrame() {
	if (!paused)
	{
		/** first preFrame ever **/
		if(mLastTiming.secd()==0)
		{
			mLastTiming = mHead->getTimeStamp();
			walkingDistance = 0;
		}
		else
		{
			/** how many time since last preFrame **/
			vpr::Interval currentTiming = mHead->getTimeStamp();
			vpr::Interval diff = currentTiming - mLastTiming;
			mLastTiming = currentTiming;
	
			/** update train position **/
			double v = speed->getSpeed(walkingDistance);
			double dt = diff.secd();
			double totalLength = path.getTotalLength();
			walkingDistance += v*dt;
			if(walkingDistance >= totalLength) walkingDistance -= totalLength;
#if 0
			printf("%f\n", 1/dt);
#endif

#if SOUND
			snx::sonix::instance()->step(dt);

			static double lastV=0;
			if(lastV)
			{
				if((v > lastV && v<0.6 && walkingDistance>3) ||
				   (v>4 && rand()<0.15*RAND_MAX))
					soundSpeed.trigger();
			}
			lastV = v;
#endif
		}

		train.setPosition(path, walkingDistance);
		camera.setFirstPersonPosition(path, walkingDistance);

		osg::Matrix mat; set(mat, mWand->getData());
		cursor.setMatrix(mat);
	}

	// first button will add/move a control point
	switch (mButton0->getData())
	{
		case gadget::Digital::OFF: break;
		case gadget::Digital::ON: break;

		case gadget::Digital::TOGGLE_OFF:
		{
			mode = NONE;
			//structure.showControlPoints(); break;
			break;
		}

		case gadget::Digital::TOGGLE_ON:
		{
			if(mode != INSERT_CONTROL_POINT && mode != MOVE_CONTROL_POINT)
				mode = INSERT_CONTROL_POINT;
			//structure.hideControlPoints(); break;
			break;
		}
	}

	// second button will insert/remove a new point
	switch (mButton1->getData())
	{
		case gadget::Digital::OFF: break;
		case gadget::Digital::ON: break;

		case gadget::Digital::TOGGLE_OFF:
		{
			mode = NONE;
			break;
		}

		case gadget::Digital::TOGGLE_ON:
		{
			float dist;
			selectedControlPointId = cursor.pick(path, dist);

			// if distance is very close to an existing control point, remove it
			if(dist<20 && mode != REMOVE_CONTROL_POINT)
				mode = REMOVE_CONTROL_POINT;
			else if(dist>=20 && mode != MOVE_WORLD)
				mode = MOVE_WORLD;
			break;
		}
	}

	// third button will put you in the riding mode
	switch (mButton2->getData())
	{
		case gadget::Digital::OFF:		break;
		case gadget::Digital::ON:		break;
		case gadget::Digital::TOGGLE_OFF:
		{
			int mode = camera.toggleMode();
			if(mode == FIRST_PERSON_CAMERA_MODE)
			{
				cursor.hide();
				structure.hideControlPoints();
			}
			else
			{
				cursor.show();
				structure.showControlPoints();
			}
			break;
		}
		case gadget::Digital::TOGGLE_ON:	break;
	}

	bool updatePath = true;

	switch(mode)
	{
	case MOVE_CONTROL_POINT:
		{
			float x, y, z;
			cursor.getTranslate(x,y,z);

			path.changeControlPoint(selectedControlPointId, x,y,z);
			break;
		}
	case INSERT_CONTROL_POINT:
		{
			float dist;
			selectedControlPointId = cursor.pick(path, dist);

			// if distance is very close to an existing control point, remove it
			if(dist < 20)
			{
				mode = MOVE_CONTROL_POINT;
			}
			// or else, add a new one
			else
			{
				float x, y, z;
				cursor.getTranslate(x,y,z);

				selectedControlPointId = path.insertControlPoint(x,y,z);
				mode = MOVE_CONTROL_POINT;
			}
			break;
		}
	case REMOVE_CONTROL_POINT:
		{
			float dist;
			selectedControlPointId = cursor.pick(path, dist);

			path.removeControlPoint(selectedControlPointId);
			mode = NONE;
			break;
		}
	case MOVE_WORLD:
		{
			osg::Matrix wand;
			set(wand, mWand->getData());
			
			float x = wand[1][0];
			float y = -wand[1][2];

			float n = x*x+y*y;

			if(n > 0.3*0.3)
				camera.translate(0.8*(x<0?1:-1),
						 0.8*(y>0?1:-1));

			break;
		}
	default:
		updatePath = false;
		break;
	}

	if(updatePath)
	{
		structure.setPath(path);
		speed->initialize();

		path.save("current.path");
	}
}


////////////////////////////////////////////////////////////////////////////////
//
// RC::initLight
//
// Cree un eclairage OpenSG. Retourne le pointeur au noeud d'eclairage,
// destine a etre ajoute au haut du graphe de scene. Utilisee par init().
// Il s'agit d'une lumiere directionelle placee au haut de la scene.
//
////////////////////////////////////////////////////////////////////////////////


osg::NodePtr RC::initLight()
{
	osg::NodePtr lightNode  = osg::Node::create();
	osg::NodePtr mLightBeacon = osg::Node::create();
	osg::TransformPtr light_beacon_core = osg::Transform::create();

	osg::Matrix light_pos;
	light_pos.setTransform(osg::Vec3f(0, 100, 0));

	osg::beginEditCP(light_beacon_core);
	light_beacon_core->setMatrix(light_pos);
	osg::endEditCP(light_beacon_core);

	osg::beginEditCP(mLightBeacon);
	mLightBeacon->setCore(light_beacon_core);
	osg::endEditCP(mLightBeacon);

	osg::DirectionalLightPtr light_core = osg::DirectionalLight::create();
	osg::beginEditCP(light_core);
	light_core->setAmbient   (0.9, 0.9, 0.9, 1);
	light_core->setDiffuse   (0.6, 0.6, 0.6, 1);
	light_core->setSpecular  (1, 1, 1, 1);
	light_core->setDirection (0, 0, 1);
	light_core->setBeacon    (lightNode);
	osg::endEditCP(light_core);

	osg::beginEditCP(lightNode);
	lightNode->setCore(light_core);
	lightNode->addChild(mLightBeacon);
	osg::endEditCP(lightNode);

	return lightNode;
}


////////////////////////////////////////////////////////////////////////////////
//
// RC::initHouse()
//
// Retourne un noeud OpenSG contenant la geometrie d'un batiment. Destine a
// etre ajoute au graphe de scene. Utilisee par init(). Le batiment est un
// bloc.
//
////////////////////////////////////////////////////////////////////////////////


osg::NodePtr RC::initHouse()
{
	// Create Geometry
        osg::SimpleMaterialPtr material = osg::SimpleMaterial::create();
	osg::beginEditCP(material);
	  material->setAmbient(osg::Color3f(0.1, 0.1, 0.3));
	  material->setDiffuse(osg::Color3f(0.1, 0.1, 0.3));
	osg::endEditCP(material);

	osg::GeometryPtr geo = osg::makeBoxGeo(6.0, 4.0, 15.0, 1, 1, 1);
	osg::beginEditCP(geo);
	  geo->setMaterial(material);
	osg::endEditCP(geo);

	// Geometry node
	osg::NodePtr geometry = osg::Node::create();
	osg::beginEditCP(geometry);
	  geometry->setCore(geo);
	osg::endEditCP(geometry);

	// Create transformation
	osg::TransformPtr t = osg::Transform::create();
	osg::Matrix m; m.setTranslate(-12,4,0);
	t->setMatrix(m);

	// Transformation node
	osg::NodePtr transf = osg::Node::create();
	beginEditCP(transf);
	  transf->setCore(t);
	  transf->addChild(geometry);
	endEditCP(transf);

	return transf;
}


////////////////////////////////////////////////////////////////////////////////
//
// RC::initSky()
//
// Retourne un noeud OpenSG contenant le ciel. Destine a etre ajoute au
// graphe de scene. Utilisee par init(). Le ciel est composee d'une grosse
// boite qui recouvre toute la scene et qui est recouverte d'une texture de
// ciel.
//
////////////////////////////////////////////////////////////////////////////////


osg::NodePtr RC::initSky()
{
#if TEXTURE
	// Texture
	osg::ImagePtr img = osg::Image::create();
	img->read("images/clouds.png");

	// Green material
	osg::SimpleTexturedMaterialPtr material = osg::SimpleTexturedMaterial::create();
	beginEditCP(material);
	  material->setImage(img);
	endEditCP(material);
#else
	osg::SimpleMaterialPtr material = osg::SimpleMaterial::create();
	osg::beginEditCP(material);
	  material->setAmbient(osg::Color3f(0.1, 0.2, 0.3));
	  material->setDiffuse(osg::Color3f(0.1, 0.2, 0.3));
	osg::endEditCP(material);
#endif

	osg::GeometryPtr geo = osg::makeBoxGeo(1.0, 1.0, 1.0, 1, 1, 1);
	osg::beginEditCP(geo);
	  geo->setMaterial(material);
	osg::endEditCP(geo);

	// Geometry node
	osg::NodePtr geometry = osg::Node::create();
	osg::beginEditCP(geometry);
	  geometry->setCore(geo);
	osg::endEditCP(geometry);

	// Create transformation
	osg::TransformPtr t = osg::Transform::create();
	osg::Matrix m; m.setScale(500);
	t->setMatrix(m);

	// Transformation node
	osg::NodePtr transf = osg::Node::create();
	beginEditCP(transf);
	  transf->setCore(t);
	  transf->addChild(geometry);
	endEditCP(transf);

	return transf;
}
