/**
 * @file speed.cpp
 * @author Herve Lombaert
 * @date October 23rd, 2005
 *
 * Given a path, it determines the speed on this path
 **/

#include "speed.h"
#include "path.h"

const double g = 9.81;
const double distanceScale = 0.15; /** units are arbitrary, how many units we have per meter **/
const double vh = 0.5f; /** speed at highest position **/
const double heightScale = 1.0f; /** maximum height is multiplied by this number **/

/** @brief constructor **/
Speed::Speed(const Path* path)
:	mPath(path),
	mSetDistance(0),
	mUnsetDistance(0),
	mMaxHeight(0),
	mV0(0),
	mV1(0)
{
	initialize();
}

/** @brief gives the speed on a position on path **/
double Speed::getSpeed(float distance) const
{
	double speed = 0;

	/** if accelerating at the beginning **/
#if 1
	if(distance < 0.25*mSetDistance)
	{
		const float ratio = distance / (mSetDistance/4);
		speed = (mV0*0.7)*ratio+0.3*mV0;
	}
	else
#endif
	if(distance < mSetDistance)
	{
		speed = mV0;
	}

	/** if in the ride **/
	else if(distance < mUnsetDistance)
	{
		float matrix[16];
		mPath->interpolate(distance, matrix);
		speed = speed2(matrix[14]);
	}

	/** if in the decceleration zone **/
	else
	{
		const float maxDistance = mPath->getTotalLength();
		const float ratio = (maxDistance-distance)/(maxDistance-mUnsetDistance);
		speed = (mV1*0.9)*ratio+0.1*mV1;
	}

	return speed;
}

/** @brief speed
    @note  cinetic energy:	ec = 1/2 mv^2
	   potential energy:	ep = mgh, where at h_max, v = vh
	   total energy:	ec0+ep0 = ec1+ecp1
	   speed if then:
		1/2 mv^2 + mgh = 1/2 m(vh)^2 + mg(h_max)
		v^2 + 2gh = (vh)^2 + 2g(h_max)
		v^2 = (vh)^2 + 2g(h_max-h)
 **/
inline double Speed::speed2(float height) const
{
	return sqrt(2*g*distanceScale*(heightScale*mMaxHeight-height)+vh*vh);
}

/** @brief initialize parameters **/
void Speed::initialize()
{
	/** set constants **/
	float dist = 0;
	float maxDistance = mPath->getTotalLength();
	const float step = 0.1f;
	const float minSetDistance = 5.0f;
	mMaxHeight = 0;

	/** search for first fall **/
	for(dist=minSetDistance; dist<maxDistance; dist+=step)
	{
		float matrix[2][16];
		mPath->interpolate(dist-step, matrix[0]);
		mPath->interpolate(dist, matrix[1]);

		if(matrix[1][14]>mMaxHeight)
			mMaxHeight = matrix[1][14];

		float deltaHeight = matrix[1][14] - matrix[0][14];
		if(deltaHeight<0)
			break;
	}
	mSetDistance = dist;

	/** what's the maximum height **/
	for(dist = mSetDistance+step; dist<maxDistance; dist+=step)
	{
		float matrix[16];
		mPath->interpolate(dist, matrix);
		if(matrix[14]>mMaxHeight)
			mMaxHeight = matrix[14];
	}

	/** make sure speed does not zero at higher point **/
	//mMaxHeight *= 2.00;

	/** set unset distance **/
	mUnsetDistance = maxDistance-mSetDistance;

	/** initial and final speed of free ride **/
	float matrix[16];
	mPath->interpolate(mSetDistance, matrix);
	mV0 = speed2(matrix[14]);
	mPath->interpolate(mUnsetDistance, matrix);
	mV1 = speed2(matrix[14]);
}
