////////////////////////////////////////////////////////////////////////////////
//
// structure.cpp
// Methodes de la classe Structure
//
// Editeur de roller coaster
// Projet de realite virtuelle (INF6802)
// Herve Lombaert et Yves Chouinard
// Automne 2005
//
////////////////////////////////////////////////////////////////////////////////


#include <OpenSG/OSGGroup.h>
#include <OpenSG/OSGTransform.h>
#include <OpenSG/OSGSwitch.h>

#include "structure.h"


#define RAIL_SECTION_LENGTH 0.4


////////////////////////////////////////////////////////////////////////////////
//
// Structure::init()
//
// Construit la base du graphe de la structure, ainsi que les materiaux et
// geometries qui ne sont plus modifies par la suite. Retourne le noeud
// OpenSG destine a etre ajoute au graphe de scene.
//
////////////////////////////////////////////////////////////////////////////////


osg::NodePtr Structure::init()
{
	//
	// Empty root nodes; controlPointRootNode and structureRootNode are
	// both children of rootNode; the control points are kept in a separate
	// structure for selection purposes
	//

	// Control point base node - child of root node
	controlPointRootNode = osg::Node::create();
	osg::beginEditCP(controlPointRootNode);
	controlPointRootNode->setCore(osg::Group::create());
	osg::addRefCP(controlPointRootNode);
	osg::endEditCP(controlPointRootNode);

	// Structure node - child of root node
	structureRootNode = osg::Node::create();
	osg::beginEditCP(structureRootNode);
	structureRootNode->setCore(osg::Group::create());
	osg::addRefCP(structureRootNode);
	osg::endEditCP(structureRootNode);

	// Root node
	rootNode = osg::Node::create();
	osg::beginEditCP(rootNode);
	rootNode->setCore(osg::Group::create());
	rootNode->addChild(controlPointRootNode);
	rootNode->addChild(structureRootNode);
	osg::addRefCP(rootNode);
	osg::endEditCP(rootNode);

	// Invisible material for hidden control points
        invisibleMaterial = osg::SimpleMaterial::create();
	osg::beginEditCP(invisibleMaterial);
	invisibleMaterial->setTransparency(1.0);
	osg::addRefCP(invisibleMaterial);
	osg::endEditCP(invisibleMaterial);

	// Normal control point material (semi-transparent green)
        normalControlPointMaterial = osg::SimpleMaterial::create();
	osg::beginEditCP(normalControlPointMaterial);
	normalControlPointMaterial->setAmbient(osg::Color3f(0.2, 0.4, 0.2));
	normalControlPointMaterial->setDiffuse(osg::Color3f(0.0, 1.0, 0.0));
	normalControlPointMaterial->setTransparency(0.5);
	osg::addRefCP(normalControlPointMaterial);
	osg::endEditCP(normalControlPointMaterial);

	// Selected control point material (semi-transparent yellow)
        selectedControlPointMaterial = osg::SimpleMaterial::create();
	osg::beginEditCP(selectedControlPointMaterial);
	selectedControlPointMaterial->setAmbient(osg::Color3f(0.4, 0.4, 0.2));
	selectedControlPointMaterial->setDiffuse(osg::Color3f(1.0, 1.0, 0.0));
	selectedControlPointMaterial->setTransparency(0.5);
	osg::addRefCP(selectedControlPointMaterial);
	osg::endEditCP(selectedControlPointMaterial);

	// Structure material (light gray)
        osg::SimpleMaterialPtr structureMaterial = osg::SimpleMaterial::create();
	osg::beginEditCP(structureMaterial);
	structureMaterial->setAmbient(osg::Color3f(0.2, 0.2, 0.2));
	structureMaterial->setDiffuse(osg::Color3f(0.5, 0.3, 0.3));
	structureMaterial->setSpecular(osg::Color3f(0.5, 0.5, 0.7));
	osg::addRefCP(structureMaterial);
	osg::endEditCP(structureMaterial);

	// Control point geometry (box)
	controlPointGeometry = osg::makeBoxGeo(1.0, 1.0, 1.0, 1, 1, 1);
	osg::beginEditCP(controlPointGeometry);
	controlPointGeometry->setMaterial(normalControlPointMaterial);
	osg::addRefCP(controlPointGeometry);
	osg::endEditCP(controlPointGeometry);

	// Rail section geometry
	railSectionGeometry = osg::makeCylinderGeo(RAIL_SECTION_LENGTH * 1.4, 0.1, 3, true, false, false);
	osg::beginEditCP(railSectionGeometry);
	railSectionGeometry->setMaterial(structureMaterial);
	osg::addRefCP(railSectionGeometry);
	osg::endEditCP(railSectionGeometry);

	return (rootNode);
}


////////////////////////////////////////////////////////////////////////////////
//
// Structure::setPath()
//
// Construit ou reconstruit toute la structure en fonction du parcours passe
// en parametre. Chaque point de controle est represente par un cube vert
// transparent et les deux rails sont composes d'une suite de petits
// cylindres.
//
////////////////////////////////////////////////////////////////////////////////


void Structure::setPath(Path path)
{
	// Destruction des anciens graphe de points de contrle et de structure
	while(controlPointRootNode->getNChildren()>0)
	{
		osg::NodePtr n = controlPointRootNode->getChild(0);
		beginEditCP(controlPointRootNode, osg::Node::ChildrenFieldMask);
			controlPointRootNode->subChild(n);
		endEditCP(controlPointRootNode, osg::Node::ChildrenFieldMask);
	}

	while(structureRootNode->getNChildren()>0)
	{
		osg::NodePtr n = structureRootNode->getChild(0);
		beginEditCP(structureRootNode, osg::Node::ChildrenFieldMask);
			structureRootNode->subChild(n);
		endEditCP(structureRootNode, osg::Node::ChildrenFieldMask);
	}

	//
	// One node for each control point
	//

	for (Path::PointID iter = path.begin(); iter < path.end(); iter++)
	{
		float m[16];
		path.getControlPoint(iter, m);

		osg::Matrix matrix;
		matrix.setValue(m);

		addNode(controlPointRootNode, matrix, controlPointGeometry);
	}

	//
	// One node for each rail section; rail section length is chosen
	// as a compromise between smoothness and lightness
	//

	float totalLength = path.getTotalLength();

	for (float length = 0; length < totalLength; length += RAIL_SECTION_LENGTH)
	{
		osg::Matrix matrix, relativeMatrix;

		float m[16];
		path.interpolate(length, m);

		// Left rail section
		matrix.setValue(m);
		relativeMatrix.setIdentity();
		relativeMatrix.setTranslate(osg::Vec3f(-0.4, 0.0, 0.0));
		matrix.mult(relativeMatrix);
		addNode(structureRootNode, matrix, railSectionGeometry);

		// Right rail section
		matrix.setValue(m);
		relativeMatrix.setIdentity();
		relativeMatrix.setTranslate(osg::Vec3f(+0.4, 0.0, 0.0));
		matrix.mult(relativeMatrix);
		addNode(structureRootNode, matrix, railSectionGeometry);

	}

	return;
}


////////////////////////////////////////////////////////////////////////////////
//
// Structure::showControlPoints()
// Structure::hideControlPoints()
//
// Pour rendre les points de controle visibles ou non. Normalement, on veut
// qu'ils soient visibles en mode d'edition et invisibles en mode simulation.
//
////////////////////////////////////////////////////////////////////////////////


void Structure::showControlPoints()
{
	osg::beginEditCP(controlPointGeometry);
	controlPointGeometry->setMaterial(normalControlPointMaterial);
	osg::endEditCP(controlPointGeometry);
}


void Structure::hideControlPoints()
{
	osg::beginEditCP(controlPointGeometry);
	controlPointGeometry->setMaterial(invisibleMaterial);
	osg::endEditCP(controlPointGeometry);
}


////////////////////////////////////////////////////////////////////////////////
//
// Structure::addNode()
//
// Methode utilitaire appelee lors de la creation du graphe de la structure
// par setPath(). Ajoute comme enfant au noeud nodeToAttachTo un noeud de
// transformation cree a partir de transformMatrix et un noeud de geometrie
// compose a partir de objectGeometry.
//
////////////////////////////////////////////////////////////////////////////////


void Structure::addNode(osg::NodePtr nodeToAttachTo, osg::Matrix transformMatrix, osg::GeometryPtr objectGeometry)
{
	osg::NodePtr geometryNode = osg::Node::create();
	osg::beginEditCP(geometryNode);
	geometryNode->setCore(objectGeometry);
	osg::endEditCP(geometryNode);

	osg::TransformPtr transform = osg::Transform::create();
	osg::beginEditCP(transform);
	transform->setMatrix(transformMatrix);
	osg::endEditCP(transform);

	osg::NodePtr transformNode = osg::Node::create();
	osg::beginEditCP(transformNode);
	transformNode->setCore(transform);
	transformNode->addChild(geometryNode);
	osg::endEditCP(transformNode);

	osg::beginEditCP(nodeToAttachTo);
	nodeToAttachTo->addChild(transformNode);
	osg::endEditCP(nodeToAttachTo);
}
