////////////////////////////////////////////////////////////////////////////////
//
// train.cpp
// Methodes des classes Train et Wagon
//
// Editeur de roller coaster
// Projet de realite virtuelle (INF6802)
// Herve Lombaert et Yves Chouinard
// Automne 2005
//
////////////////////////////////////////////////////////////////////////////////


#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGSimpleMaterial.h>

#include "train.h"


////////////////////////////////////////////////////////////////////////////////
//
// Wagon::init()
//
// Cree et retourne le noeud OpenSG d'un wagon du train. Destine a etre
// ajoute au graphe du train par Train::init(). Comme chaque wagon est a une
// distance fixe du rail, on definit tout de suite un noeud de transformation
// constant pour exprimer cette distance.
//
// Le wagon est un bloc rouge.
//
////////////////////////////////////////////////////////////////////////////////


osg::NodePtr Wagon::init()
{
	// Red material
        osg::SimpleMaterialPtr material = osg::SimpleMaterial::create();
	osg::beginEditCP(material);
	material->setAmbient(osg::Color3f(0.2, 0.1, 0.1));
	material->setDiffuse(osg::Color3f(0.5, 0.1, 0.1));
	material->setSpecular(osg::Color3f(0.8, 0.0, 0.0));
	osg::endEditCP(material);

	// Wagon geometry - box for now - x left (right?), y forward, z up
	osg::GeometryPtr geometry = osg::makeBoxGeo(0.6, 1.0, 0.4, 1, 1, 1);
	osg::beginEditCP(geometry);
	geometry->setMaterial(material);
	osg::endEditCP(geometry);

	// Geometry node
	osg::NodePtr geometryNode = osg::Node::create();
	osg::beginEditCP(geometryNode);
	geometryNode->setCore(geometry);
	osg::endEditCP(geometryNode);

	// Transform - position relative to rail
	osg::Matrix matrix;
	matrix.setTranslate(0, 0, 0.2);

	osg::TransformPtr relativeTransform = osg::Transform::create();
	osg::beginEditCP(relativeTransform);
	relativeTransform->setMatrix(matrix);
	osg::endEditCP(relativeTransform);

	osg::NodePtr relativeTransformNode = osg::Node::create();
	osg::beginEditCP(relativeTransformNode);
	relativeTransformNode->setCore(relativeTransform);
	relativeTransformNode->addChild(geometryNode);
	osg::endEditCP(relativeTransformNode);

	// Transform - position on path - matrix first assigned by setTransform()
	transform = osg::Transform::create();
	osg::addRefCP(transform);

	osg::NodePtr transformNode = osg::Node::create();
	osg::beginEditCP(transformNode);
	transformNode->setCore(transform);
	transformNode->addChild(relativeTransformNode);
	osg::addRefCP(transformNode);
	osg::endEditCP(transformNode);

	return transformNode;
}


////////////////////////////////////////////////////////////////////////////////
//
// Wagon::setTransform()
//
// Met a jour la position du wagon en donnant directement sa nouvelle matrice
// de transformation.
//
////////////////////////////////////////////////////////////////////////////////


void Wagon::setTransform(float matrix[16])
{
	osg::Matrix m;
	m.setValue(matrix);

	osg::beginEditCP(transform);
	transform->setMatrix(m);
	osg::endEditCP(transform);
}


////////////////////////////////////////////////////////////////////////////////
//
// Train::init()
//
// Cree le train comme un ensemble de wagons. Chacun des wagons est un objet
// de la classe Wagon et est cree par Wagon::init(). Retourne le noeud OpenSG
// contenant le sous-graphe du train. Pour etre ajoute au graphe de scene.
//
////////////////////////////////////////////////////////////////////////////////


osg::NodePtr Train::init(unsigned int nbWagons)
{
	osg::NodePtr trainNode = osg::Node::create();
	beginEditCP(trainNode);
	trainNode->setCore(osg::Group::create());
	endEditCP(trainNode);

	for (unsigned int i = 0; i < nbWagons; i++)
	{
		Wagon wagon;

		osg::beginEditCP(trainNode);
		trainNode->addChild(wagon.init());
		osg::endEditCP(trainNode);

		wagons.push_back(wagon);
	}

	return trainNode;
}


////////////////////////////////////////////////////////////////////////////////
//
// Train::setPosition()
//
// Pour mettre a jour la position du train sur les rails. Le parcours est
// specifie par path et la position sur le parcours par walkingDistance. Les
// wagons sont places a intervalles reguliers sur le parcours. La matrice de
// transformation de chacun est calculee et mise a jour.
//
////////////////////////////////////////////////////////////////////////////////


void Train::setPosition(Path &path, float walkingDistance)
{
	float distanceBetweenWagons = 3;
	int nbWagons = wagons.size();
	float positionOfWagon = walkingDistance + (nbWagons/2-1)*distanceBetweenWagons;
	float totalLength = path.getTotalLength();

	try
	{

	for (std::vector<Wagon>::iterator wagon = wagons.begin(); wagon != wagons.end(); wagon++)
	{
		if (positionOfWagon > totalLength) positionOfWagon -= totalLength;
		if (positionOfWagon < 0) positionOfWagon += totalLength;

		float matrix[16];
		path.interpolate(positionOfWagon, matrix);
		wagon->setTransform(matrix);

		positionOfWagon -= distanceBetweenWagons;
	}

	} catch(Path::Exception &e) {
		e.what();
		printf("positionOfFrontWagon: %f (%f)\n", positionOfWagon, totalLength);
		throw e;
	}
}
